import datetime
import sys
from abc import abstractmethod

import six
import yaml
from decorator import decorator

from omniduct.duct import Duct
from omniduct.utils.config import config
from omniduct.utils.debug import logger
from omniduct.utils.decorators import function_args_as_kwargs
from omniduct.utils.docs import quirk_docs

from ._serializers import PickleSerializer

config.register(
    'cache_fail_hard',
    description='Raise an exception if a cache fails to save (otherwise errors are logged and suppressed).',
    default=False
)


def cached_method(
        key,
        namespace=lambda self, kwargs: (
            self.cache_namespace or "{}.{}".format(self.__class__.__name__, self.name)
        ),
        cache=lambda self, kwargs: self.cache,
        use_cache=lambda self, kwargs: kwargs.pop('use_cache', True),
        renew=lambda self, kwargs: kwargs.pop('renew', False),
        serializer=lambda self, kwargs: PickleSerializer(),
        metadata=lambda self, kwargs: None
):
    """
    Wrap a method of a `Duct` class and add caching capabilities.

    All arguments of this function are expected to be functions taking two
    arguments: a reference to current instance of the class (`self`) and a
    dictionary of arguments passed to the function (`kwargs`).

    Args:
        key (function -> str): The key under which the value returned by the
            wrapped function should be stored.
        namespace (function -> str): The namespace under which the key should be
            stored (default: `"<duct class name>.<duct instance name>"`).
        cache (function -> Cache): The instance of cache via which to store the
            output of the wrapped function (default: `self.cache`).
        use_cache (function -> bool): Whether or not to use the caching
            functionality (default: `True`).
        renew (function -> bool): Whether to renew the stored cache, overriding
            if a value has already been stored (default: `False`).
        serializer (function -> Serializer): The `Serializer` subclass to use
            when storing the return object (default: `PickleSerializer`).
        metadata (function -> None, dict): A dictionary of additional metadata
            to be stored alongside the wrapped function's output
            (default: `None`).

    Returns:
        object: The (potentially cached) object returned when calling the
            wrapped function.

    Raises:
        Exception: If cache fails to store the output of the wrapped function,
            and the omniduct configuration key `cache_fail_hard` is `True`, then
            the underlying exceptions raised by the Cache instance will be
            reraised.
    """

    @decorator
    def wrapped(method, self, *args, **kwargs):
        kwargs = function_args_as_kwargs(method, self, *args, **kwargs)
        kwargs.pop('self')

        _key = key(self, kwargs)
        _namespace = namespace(self, kwargs)
        _cache = cache(self, kwargs)
        _use_cache = use_cache(self, kwargs)
        _renew = renew(self, kwargs)
        _serializer = serializer(self, kwargs)
        _metadata = metadata(self, kwargs)

        if _cache is None or not _use_cache:
            return method(self, **kwargs)

        if _renew or not _cache.has_key(_key, namespace=_namespace):  # noqa: has_key is not of a dictionary here
            value = method(self, **kwargs)
            if value is None:
                logger.warning("Method value returned None. Not saving to cache.")
                return

            try:
                _cache.set(
                    _key,
                    value=value,
                    namespace=_namespace,
                    serializer=_serializer,
                    metadata=_metadata
                )
            except:
                logger.warning("Failed to save results to cache. If needed, please save them manually.")
                if config.cache_fail_hard:
                    six.reraise(*sys.exc_info())
        else:
            logger.caveat('Loaded from cache')

        # Return from cache every time, just in case serialization operation was
        # destructive (e.g. reading from cursors)
        return _cache.get(
            _key,
            namespace=_namespace,
            serializer=_serializer
        )
    return wrapped


class Cache(Duct):
    """
    An abstract class providing the common API for all cache clients.
    """

    DUCT_TYPE = Duct.Type.CACHE

    @quirk_docs('_init', mro=True)
    def __init__(self, **kwargs):
        Duct.__init_with_kwargs__(self, kwargs)
        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    # Data insertion and retrieval

    def set(self, key, value, namespace=None, serializer=None, metadata=None):
        """
        Set the value of a key.

        Args:
            key (str): The key for which `value` should be stored.
            value (object): The value to be stored.
            namespace (str, None): The namespace to be used.
            serializer (Serializer): The `Serializer` subclass to use for the
                serialisation of value into the cache. (default=PickleSerializer)
            metadata (dict, None): Additional metadata to be stored with the value
                in the cache. Values must be serializable via `yaml.safe_dump`.
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        serializer = serializer or PickleSerializer()
        try:
            with self._get_stream_for_key(namespace, key, 'data{}'.format(serializer.file_extension), mode='wb', create=True) as fh:
                serializer.serialize(value, fh)
            self.set_metadata(key, metadata, namespace=namespace, replace=True)
        except:
            self.unset(key, namespace=namespace)
            six.reraise(*sys.exc_info())

    def set_metadata(self, key, metadata, namespace=None, replace=False):
        """
        Set the metadata associated with a stored key, creating the key if it
        is missing.

        Args:
            key (str): The key for which `value` should be stored.
            metadata (dict, None): Additional/override metadata to be stored
                for `key` in the cache. Values must be serializable via
                `yaml.safe_dump`.
            namespace (str, None): The namespace to be used.
            replace (bool): Whether the provided metadata should entirely
                replace any existing metadata, or just update it. (default=False)
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        if replace:
            orig_metadata = {'created': datetime.datetime.utcnow()}
        else:
            orig_metadata = self.get_metadata(key, namespace=namespace)

        orig_metadata.update(metadata or {})

        with self._get_stream_for_key(namespace, key, 'metadata', mode='w', create=True) as fh:
            yaml.safe_dump(orig_metadata, fh, default_flow_style=False)

    def get(self, key, namespace=None, serializer=None):
        """
        Retrieve the value associated with the nominated key from the cache.

        Args:
            key (str): The key for which `value` should be retrieved.
            namespace (str, None): The namespace to be used.
            serializer (Serializer): The `Serializer` subclass to use for the
                deserialisation of value from the cache. (default=PickleSerializer)

        Returns:
            object: The (appropriately deserialized) object stored in the cache.
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        serializer = serializer or PickleSerializer()
        if not self._has_key(namespace, key):
            raise KeyError("{} (namespace: {})".format(key, namespace))
        try:
            with self._get_stream_for_key(namespace, key, 'data{}'.format(serializer.file_extension), mode='rb', create=False) as fh:
                return serializer.deserialize(fh)
        finally:
            self.set_metadata(key, namespace=namespace, metadata={'last_accessed': datetime.datetime.utcnow()})

    def get_metadata(self, key, namespace=None):
        """
        Retrieve metadata associated with the nominated key from the cache.

        Args:
            key (str): The key for which to extract metadata.
            namespace (str, None): The namespace to be used.

        Returns:
            dict: The metadata associated with this namespace and key.
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        if not self._has_key(namespace, key):
            raise KeyError("{} (namespace: {})".format(key, namespace))
        try:
            with self._get_stream_for_key(namespace, key, 'metadata', mode='r', create=False) as fh:
                return yaml.safe_load(fh)
        except:
            return {}

    def unset(self, key, namespace=None):
        """
        Remove the nominated key from the cache.

        Args:
            key (str): The key which should be unset.
            namespace (str, None): The namespace to be used.
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        if not self._has_key(namespace, key):
            raise KeyError("{} (namespace: {})".format(key, namespace))
        self._remove_key(namespace, key)

    def unset_namespace(self, namespace=None):
        """
        Remove an entire namespace from the cache.

        Args:
            namespace (str, None): The namespace to be removed.
        """
        self.connect()
        namespace = self._namespace(namespace)
        if not self._has_namespace(namespace):
            raise KeyError("namespace: {}".format(namespace))
        self._remove_namespace(namespace)

    # Top-level descriptions

    @property
    def namespaces(self):
        "list <str,None>: A list of the namespaces stored in the cache."
        return self.connect()._get_namespaces()

    def has_namespace(self, namespace=None):
        """
        Check whether the cache has the nominated namespace.

        Args:
            namespace (str,None): The namespace for which to check for existence.

        Returns:
            bool: Whether the cache has the nominated namespaces.
        """
        self.connect()
        namespace = self._namespace(namespace)
        return self._has_namespace(namespace)

    def keys(self, namespace=None):
        """
        Collect a list of all the keys present in the nominated namespaces.

        Args:
            namespace (str,None): The namespace from which to extract all of the
                keys.

        Returns:
            list<str>: The keys stored in the cache for the nominated namespace.
        """
        self.connect()
        namespace = self._namespace(namespace)
        return self._get_keys(namespace)

    def has_key(self, key, namespace=None):
        """
        Check whether the cache as a nominated key.

        Args:
            key (str): The key for which to check existence.
            namespace (str,None): The namespace from which to extract all of the
                keys.

        Returns:
            bool: Whether the cache has a value for the nominated namespace and
                key.
        """
        self.connect()
        namespace, key = self._namespace(namespace), self._key(key)
        return self._has_key(namespace, key)

    # Methods for subclasses to implement

    def _namespace(self, namespace):
        return namespace

    def _key(self, key):
        return key

    @abstractmethod
    def _get_namespaces(self):
        raise NotImplementedError

    def _has_namespace(self, namespace):
        return namespace in self._get_namespaces()

    @abstractmethod
    def _remove_namespace(self, namespace):
        raise NotImplementedError

    @abstractmethod
    def _get_keys(self, namespace):
        raise NotImplementedError

    def _has_key(self, namespace, key):
        return key in self._get_keys(namespace=namespace)

    @abstractmethod
    def _remove_key(self, namespace, key):
        raise NotImplementedError

    @abstractmethod
    def _get_stream_for_key(self, namespace, key, stream_name, mode, create):
        pass



import inspect
import os
import six
import textwrap
from abc import ABCMeta

import decorator

from omniduct.errors import DuctProtocolUnknown
from omniduct.utils.debug import logger


class ProtocolRegisteringABCMeta(ABCMeta):
    """
    This metaclass provides automatic registration of Duct subclasses so that
    they can be looked up by the protocols they support. Note that protocol
    mappings must be unique.
    """

    def __init__(cls, name, bases, dct):
        ABCMeta.__init__(cls, name, bases, dct)

        if not hasattr(cls, '_protocols'):
            cls._protocols = {}

        registry_keys = getattr(cls, 'PROTOCOLS', []) or []
        if registry_keys:
            for key in registry_keys:
                if key in cls._protocols and cls.__name__ != cls._protocols[key].__name__:
                    logger.info("Ignoring attempt by class `{}` to register key '{}', which is already registered for class `{}`.".format(cls.__name__, key, cls._protocols[key].__name__))
                else:
                    cls._protocols[key] = cls

    def _for_protocol(cls, key):
        if key not in cls._protocols:
            raise DuctProtocolUnknown("Missing `Duct` implementation for protocol: '{}'.".format(key))
        return cls._protocols[key]


class ProtocolRegisteringQuirkDocumentedABCMeta(ProtocolRegisteringABCMeta):
    """
    This metaclass adds the ability to automatically append quirk documentation
    to methods from a nominated method. For example, if the protocol specific
    implementation of `.connect()` is implemented in `._connect`, you can
    decorate the connect method with this decorator using
    `@quirk_docs('_connect')`, then the documentation from the `_connect`
    method will be appended to the `connect` docs under a heading "<cls> Quirks:".
    """

    def __init__(cls, name, bases, dct):
        super(ProtocolRegisteringQuirkDocumentedABCMeta, cls).__init__(name, bases, dct)

        # Allow method of avoiding appending of quirk docs in some environments (such as documentation)
        if os.environ.get('OMNIDUCT_DISABLE_QUIRKDOCS', None) is not None:
            return

        @decorator.decorator
        def wrapped(f, *args, **kw):
            return f(*args, **kw)

        mro = inspect.getmro(cls)
        mro = mro[:[klass.__name__ for klass in mro].index('Duct') + 1]

        # Handle module-level documentation
        module_docs = [cls.__doc__]
        for klass in mro:
            if klass != cls and hasattr(klass, '_{}__doc_attrs'.format(klass.__name__)):
                module_docs.append([
                    'Attributes inherited from {}:'.format(klass.__name__),
                    inspect.cleandoc(getattr(klass, '_{}__doc_attrs'.format(klass.__name__)))
                ])

        cls.__doc__ = cls.__doc_join(*module_docs)

        # Handle function/method-level documentation
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):

            # Check if there is anything to do
            if (
                inspect.isabstract(member) or
                not (
                    getattr(member, '_quirks_method', None) or
                    getattr(member, '_quirks_mro', False)
                )
            ):
                continue

            local_member = name in cls.__dict__

            # Extract documentation from this member and the quirks member
            member_docs = getattr(member, '__doc_orig__', None) or getattr(member, '__doc__')
            mro_docs = quirk_docs = None
            mro_order = reversed(mro) if member._quirks_mro_reverse else mro
            if member._quirks_mro:
                mro_docs = cls.__doc_join(
                    *[
                        [
                            'Inherited via {}:'.format(klass.__name__),
                            getattr(getattr(klass, member.__name__), '__doc_orig__', None) or getattr(klass, member.__name__).__doc__
                        ]
                        for klass in mro_order if member.__name__ in klass.__dict__
                    ]
                )
            if member._quirks_method and member._quirks_method in cls.__dict__:
                quirk_member = getattr(cls, member._quirks_method, None)
                if quirk_member:
                    quirk_docs = getattr(quirk_member, '__doc_orig__', None) or getattr(quirk_member, '__doc__')

            if quirk_docs or mro_docs:
                # Overide method object with new object so we don't modify
                # underlying method that may be shared by multiple classes.
                setattr(cls, name, wrapped(member))
                member = getattr(cls, name)
                member.__doc__ = cls.__doc_join(
                    member_docs if (local_member or not mro_docs) else None,
                    mro_docs,
                    [
                        "{} Quirks:".format(cls.__name__),
                        quirk_docs
                    ]
                )

    @classmethod
    def __doc_join(cls, *docs, **kwargs):
        out = []
        for doc in docs:
            if doc in (None, ''):
                continue
            elif isinstance(doc, six.string_types):
                out.append(textwrap.dedent(doc).strip('\n'))
            elif isinstance(doc, (list, tuple)):
                if len(doc) < 2:
                    continue
                d = cls.__doc_join(*doc[1:])
                if d:
                    out.append(
                        '{header}\n{body}'.format(
                            header=doc[0].strip(),
                            body='    ' + d.replace('\n', '\n    ')  # textwrap.indent not available in python2
                        )
                    )
            else:
                raise ValueError("Unrecognised doc format: {}".format(type(doc)))
        return '\n\n'.join(out)



from omniduct.filesystems.base import FileSystemClient


class StubFsClient(FileSystemClient):

    PROTOCOLS = []
    DEFAULT_PORT = None

    def _init(self):
        pass

    # Connection management

    def _connect(self):
        raise NotImplementedError

    def _is_connected(self):
        raise NotImplementedError

    def _disconnect(self):
        raise NotImplementedError

    # Path properties and helpers

    def _path_home(self):
        return NotImplementedError

    def _path_separator(self):
        raise NotImplementedError

    # File node properties

    def _exists(self, path):
        raise NotImplementedError

    def _isdir(self, path):
        raise NotImplementedError

    # Directory handling and enumeration

    def _dir(self, path):
        raise NotImplementedError

    def _mkdir(self, path, recursive, exist_ok):
        raise NotImplementedError

    # File handling

    # Either re-implement _open, or implement the _file_*_ methods below.
    # def _open(path, mode):
    #     raise NotImplementedError

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        raise NotImplementedError

    def _file_write_(self, path, s, binary):
        raise NotImplementedError

    def _file_append_(self, path, s, binary):
        raise NotImplementedError



import inspect
import sys

import six


def function_args_as_kwargs(func, *args, **kwargs):
    if six.PY3 and not hasattr(sys, 'pypy_version_info'):
        arguments = inspect.signature(func).parameters.keys()
    else:
        arguments = inspect.getargspec(func).args
    kwargs.update(dict(zip(list(arguments), args)))
    return kwargs



import hashlib

import pandas as pd
import pytest

from omniduct.databases.base import DatabaseClient


class DummyDatabaseClient(DatabaseClient):

    PROTOCOLS = []
    DEFAULT_PORT = None

    def _init(self):
        pass

    # Connection management

    def _connect(self):
        pass

    def _is_connected(self):
        return True

    def _disconnect(self):
        pass

    # Database operations

    def _execute(self, statement, cursor, wait, session_properties, **kwargs):
        return DummyCursor()

    def _table_list(self, namespace, **kwargs):
        raise NotImplementedError

    def _table_drop(self, table, **kwargs):
        pass

    def _table_exists(self, table, **kwargs):
        raise NotImplementedError

    def _table_desc(self, table, **kwargs):
        raise NotImplementedError

    def _table_head(self, table, n=10, **kwargs):
        raise NotImplementedError

    def _table_props(self, table, **kwargs):
        raise NotImplementedError


class DummyCursor(object):
    """
    This DBAPI2 compatible cursor wrapped around a Pandas DataFrame
    """

    def __init__(self):
        self.df = pd.DataFrame(
            {'field1': list(range(10)), 'field2': list('abcdefghij')}
        )
        self._df_iter = None

    @property
    def df_iter(self):
        if not getattr(self, '_df_iter'):
            self._df_iter = (tuple(row) for i, row in self.df.iterrows())
        return self._df_iter

    arraysize = 1

    @property
    def description(self):
        return tuple([
            (name, None, None, None, None, None, None)
            for name in self.df.columns
        ])

    @property
    def row_count(self):
        return -1

    def close(self):
        pass

    def execute(operation, parameters=None):
        pass

    def executemany(operation, seq_of_parameters=None):
        pass

    def fetchone(self):
        return next(self.df_iter)

    def fetchmany(self, size=None):
        size = size or self.arraysize
        return [self.fetchone() for _ in range(size)]

    def fetchall(self):
        return list(self.df_iter)

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass


class TestDatabaseClient:

    @pytest.fixture
    def db_client(self):
        return DummyDatabaseClient()

    def test_query(self, db_client):
        result = db_client.query("DUMMY QUERY")

        assert type(result) == pd.DataFrame
        assert list(result.columns) == ['field1', 'field2']

        assert all(db_client("DUMMY_QUERY") == result)

    def test_multiple_queries(self, mocker, db_client):
        mocked = mocker.spy(db_client, "_execute")
        db_client.query("DUMMY QUERY; DUMMY_QUERY")
        assert mocked.call_count == 2

    def test_statement_hash(self, db_client):
        statement = "DUMMY QUERY"
        assert db_client.statement_hash(statement) == hashlib.sha256(statement.encode()).hexdigest()

    def test_stream(self, db_client):
        stream = db_client.stream("DUMMY QUERY")
        row = next(stream)
        assert tuple(row) == (0, 'a')
        stream.close()

    def test_format(self, db_client):
        result = db_client.query("DUMMY QUERY", format='csv')
        assert result == "field1,field2\r\n0,a\r\n1,b\r\n2,c\r\n3,d\r\n4,e\r\n5,f\r\n6,g\r\n7,h\r\n8,i\r\n9,j\r\n"



from wrapt import ObjectProxy


class NestedDictObjectProxy(ObjectProxy):

    @classmethod
    def _from_tree(cls, tree, is_flat=False, get_nesting=None):
        if None in tree and len(tree) == 1:
            return tree[None]
        return cls(tree, is_flat=is_flat, get_nesting=get_nesting)

    def __init__(self, tree, is_flat=False, get_nesting=None):
        ObjectProxy.__init__(self, tree.get(None))
        self._self_tree = tree
        self._self_tree_is_flat = is_flat
        self._self_tree_get_nesting = get_nesting

    @property
    def __tree__(self):
        def add_nested(tree, key, value):
            if self._self_tree_get_nesting is not None:
                key = self._self_tree_get_nesting(key, value)
            else:
                key = key.split('/')
            for k in key:
                if k not in tree:
                    tree[k] = {}
                tree = tree[k]
            tree[None] = value

        def flat_to_nested(tree):
            out = {}
            for k, v in tree.items():
                add_nested(out, k, v)
            return out

        tree = self._self_tree
        if self._self_tree_is_flat:
            tree = flat_to_nested(tree)

        return tree

    def __dir__(self):
        return [l for l in list(self.__tree__) + ObjectProxy.__dir__(self) if l is not None]

    def __getattr__(self, name):
        if name in self.__tree__:
            if not isinstance(self.__tree__[name], NestedDictObjectProxy):
                return NestedDictObjectProxy._from_tree(self.__tree__[name])
            return self.__tree__[name]
        if None in self.__tree__:
            return getattr(self.__tree__[None], name)
        raise AttributeError('Invalid child node.'.format(name))

    def __repr__(self):
        if None in self.__tree__:
            return self.__wrapped__.__repr__()
        return ObjectProxy.__repr__(self)



import datetime
import errno
import os
import shutil
import six
import sys
from io import open

from .base import FileSystemClient, FileSystemFileDesc


class LocalFsClient(FileSystemClient):
    """
    `LocalFsClient` is a `Duct` that implements the `FileSystemClient` common
    API, and exposes the local filesystem.

    Unlike most other filesystems, `LocalFsClient` defaults to the current
    working directory on the local machine, rather than the home directory
    as used on remote filesystems. To change this, you can always execute:
    ```
    local_fs.path_cwd = local_fs.path_home
    ```
    """

    PROTOCOLS = ['localfs']

    def _init(self):
        self._path_cwd = self._path_cwd or os.getcwd()

    def _prepare(self):
        assert self.remote is None, "LocalFsClient cannot be used in conjunction with a remote client."
        super(LocalFsClient, self)._prepare()

    def _connect(self):
        pass

    def _is_connected(self):
        return True

    def _disconnect(self):
        pass

    # File enumeration
    def _path_home(self):
        return os.path.expanduser('~')

    def _path_separator(self):
        return os.path.sep

    def _exists(self, path):
        return os.path.exists(path)

    def _isdir(self, path):
        return os.path.isdir(path)

    def _isfile(self, path):
        return os.path.isfile(path)

    def _dir(self, path):
        if not os.path.isdir(path):
            raise RuntimeError("No such folder.")
        for f in os.listdir(path):
            f_path = os.path.join(path, f)

            attrs = {}

            if os.name == 'posix':
                import grp
                import pwd

                stat = os.stat(f_path)

                attrs.update({
                    'owner': pwd.getpwuid(stat.st_uid).pw_name,
                    'group': grp.getgrgid(stat.st_gid).gr_name,
                    'permissions': oct(stat.st_mode),
                    'created': str(datetime.datetime.fromtimestamp(stat.st_ctime)),
                    'last_modified': str(datetime.datetime.fromtimestamp(stat.st_mtime)),
                    'last_accessed': str(datetime.datetime.fromtimestamp(stat.st_atime)),
                })

            yield FileSystemFileDesc(
                fs=self,
                path=f_path,
                name=f,
                type='directory' if os.path.isdir(f_path) else 'file',
                bytes=os.path.getsize(f_path),
                **attrs
            )

    def _walk(self, path):
        return os.walk(path)

    def _mkdir(self, path, recursive, exist_ok):
        try:
            os.makedirs(path) if recursive else os.mkdir(path)
        except OSError as exc:  # Python >2.5
            if exc.errno != errno.EEXIST or not exist_ok or not os.path.isdir(path):
                six.reraise(*sys.exc_info())

    def _remove(self, path, recursive):
        if recursive and self.isdir(path):
            shutil.rmtree(path)
        else:
            os.unlink(path)

    # File opening

    def _open(self, path, mode):
        return open(path, mode=mode, encoding=None if 'b' in mode else 'utf-8')



from setuptools import find_packages, setup

# Extract version information from Omniduct _version.py
version_info = {}
with open('omniduct/_version.py') as version_file:
    exec(version_file.read(), version_info)

# Extract long description from readme
with open('README.md') as readme:
    long_description = ""
    while True:
        line = readme.readline()
        if line.startswith('`omniduct`'):
            long_description = line
            break
    long_description += readme.read()

setup(
    # Package metadata
    name="omniduct",
    versioning='post',
    version=version_info['__version__'],
    author=version_info['__author__'],
    author_email=version_info['__author_email__'],
    url="https://github.com/airbnb/omniduct",
    description=(
        "A toolkit providing a uniform interface for connecting to and "
        "extracting data from a wide variety of (potentially remote) data "
        "stores (including HDFS, Hive, Presto, MySQL, etc)."
    ),
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # Package details
    packages=find_packages(),
    include_package_data=True,

    # Dependencies
    setup_requires=['setupmeta'],
    install_requires=version_info['__dependencies__'],
    extras_require=version_info['__optional_dependencies__']
)



from __future__ import absolute_import

from omniduct.utils.debug import logger

from .base import DatabaseClient
from ._schemas import SchemasMixin
from . import _pandas


class SQLAlchemyClient(DatabaseClient, SchemasMixin):
    """
    This Duct connects to several different databases using one of several
    SQLAlchemy drivers. In general, these are provided for their potential
    utility, but will be less functional than the specially crafted database
    clients.
    """

    PROTOCOLS = ['sqlalchemy', 'firebird', 'mssql', 'mysql', 'oracle', 'postgresql', 'sybase', 'snowflake']
    NAMESPACE_NAMES = ['database', 'table']
    NAMESPACE_QUOTECHAR = '"'  # TODO: Apply overrides depending on protocol?
    NAMESPACE_SEPARATOR = '.'

    def _init(self, dialect=None, driver=None, database='', engine_opts=None):

        assert self._port is not None, "Omniduct requires SQLAlchemy databases to manually specify a port, as " \
                                       "it will often be the case that ports are being forwarded."

        if self.protocol is not 'sqlalchemy':
            self.dialect = self.protocol
        else:
            self.dialect = dialect
        assert self.dialect is not None, "Dialect not specified."

        self.driver = driver
        self.database = database
        self.connection_fields += ('schema',)
        self.engine_opts = engine_opts or {}

        self.engine = None
        self.connection = None

    @property
    def db_uri(self):
        return '{dialect}://{login}@{host_port}/{database}'.format(
            dialect=self.dialect + ("+{}".format(self.driver) if self.driver else ''),
            login=self.username + (":{}".format(self.password) if self.password else ''),
            host_port=self.host + (":{}".format(self.port) if self.port else ''),
            database=self.database
        )

    def _connect(self):
        import sqlalchemy
        if self.protocol not in ['mysql']:
            logger.warning("While querying and executing should work as "
                           "expected, some operations on this database client "
                           "(such as listing tables, querying to tables, etc) "
                           "may not function as expected due to the backend "
                           "not supporting ANSI SQL.")

        self.engine = sqlalchemy.create_engine(self.db_uri, **self.engine_opts)
        self._sqlalchemy_metadata = sqlalchemy.MetaData(self.engine)

    def _is_connected(self):
        return self.engine is not None

    def _disconnect(self):
        self.engine = None
        self._sqlalchemy_metadata = None
        self._schemas = None

    def _execute(self, statement, cursor, wait, session_properties, query=True, **kwargs):
        assert wait, "`SQLAlchemyClient` does not support asynchronous operations."
        if cursor:
            cursor.execute(statement)
        else:
            cursor = self.engine.execute(statement).cursor
        return cursor

    def _query_to_table(self, statement, table, if_exists, **kwargs):
        statements = []

        if if_exists == 'fail' and self.table_exists(table):
            raise RuntimeError("Table {} already exists!".format(table))
        elif if_exists == 'replace':
            statements.append('DROP TABLE IF EXISTS {};'.format(table))
        elif if_exists == 'append':
            raise NotImplementedError("Append operations have not been implemented for {}.".format(self.__class__.__name__))

        statement = "CREATE TABLE {table} AS ({statement})".format(
            table=table,
            statement=statement
        )
        return self.execute(statement, **kwargs)

    def _dataframe_to_table(self, df, table, if_exists='fail', **kwargs):
        return _pandas.to_sql(
            df=df, name=table.table, schema=table.database, con=self.engine,
            index=False, if_exists=if_exists, **kwargs
        )

    def _table_list(self, **kwargs):
        return self.query("SHOW TABLES", **kwargs)

    def _table_exists(self, table, **kwargs):
        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except:
            return False
        finally:
            logger.disabled = False

    def _table_drop(self, table, **kwargs):
        return self.execute("DROP TABLE {table}".format(table=table))

    def _table_desc(self, table, **kwargs):
        return self.query("DESCRIBE {0}".format(table), **kwargs)

    def _table_head(self, table, n=10, **kwargs):
        return self.query("SELECT * FROM {} LIMIT {}".format(table, n), **kwargs)

    def _table_props(self, table, **kwargs):
        raise NotImplementedError



import posixpath
import random
from functools import partial

from .base import FileSystemClient, FileSystemFileDesc
from .local import LocalFsClient

# Python 2 compatibility imports
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class WebHdfsClient(FileSystemClient):
    """
    This Duct connects to an Apache WebHDFS server using the `pywebhdfs` library.

    Attributes:
        namenodes (list<str>): A list of hosts that are acting as namenodes for
            the HDFS cluster in form "<hostname>:<port>".
    """

    PROTOCOLS = ['webhdfs']
    DEFAULT_PORT = 50070

    def _init(self, namenodes=None, auto_conf=False, auto_conf_cluster=None,
              auto_conf_path=None, **kwargs):
        """
        namenodes (list<str>): A list of hosts that are acting as namenodes for
            the HDFS cluster in form "<hostname>:<port>".
        auto_conf (bool): Whether to automatically extract host, port and
            namenode information from Cloudera configuration files. If True,
            automatically extracted values will override other passed values.
        auto_conf_cluster (str): The name of the cluster for which to extract
            configuration.
        auto_conf_path (str): The path of the `hdfs-site.xml` file in which
            the HDFS configuration is stored (on the remote filesystem if
            `remote` is specified, and on the local filesystem otherwise).
            Defaults to '/etc/hadoop/conf.cloudera.hdfs2/hdfs-site.xml'.
        **kwargs (dict): Additional arguments to pass onto the WebHdfs client.
        """
        self.namenodes = namenodes

        if auto_conf:
            from ._webhdfs_helpers import CdhHdfsConfParser

            assert auto_conf_cluster is not None, "You must specify a cluster via `auto_conf_cluster` for auto-detection to work."

            def get_host_and_set_namenodes(duct, cluster, conf_path):
                conf_parser = CdhHdfsConfParser(duct.remote or LocalFsClient(), conf_path=conf_path)
                duct.namenodes = conf_parser.namenodes(cluster)
                return random.choice(duct.namenodes)

            self._host = partial(get_host_and_set_namenodes, cluster=auto_conf_cluster, conf_path=auto_conf_path)

        self.__webhdfs = None
        self.__webhdfs_kwargs = kwargs
        self.prepared_fields += ('namenodes',)

    def _connect(self):
        from ._webhdfs_helpers import OmniductPyWebHdfsClient
        self.__webhdfs = OmniductPyWebHdfsClient(
            host=self._host,
            port=self._port,
            remote=self.remote,
            namenodes=self.namenodes,
            user_name=self.username,
            **self.__webhdfs_kwargs
        )

    def _is_connected(self):
        try:
            if self.remote and not self.remote.is_connected():
                return False
            return self.__webhdfs is not None
        except:
            return False

    def _disconnect(self):
        self.__webhdfs = None

    # Path properties and helpers

    def _path_home(self):
        return self.__webhdfs.get_home_directory()

    def _path_separator(self):
        return '/'

    # File node properties

    def _exists(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            self.__webhdfs.get_file_dir_status(path)
            return True
        except FileNotFound:
            return False

    def _isdir(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            stats = self.__webhdfs.get_file_dir_status(path)
            return stats['FileStatus']['type'] == 'DIRECTORY'
        except FileNotFound:
            return False

    def _isfile(self, path):
        from pywebhdfs.errors import FileNotFound
        try:
            stats = self.__webhdfs.get_file_dir_status(path)
            return stats['FileStatus']['type'] == 'FILE'
        except FileNotFound:
            return False

    # Directory handling and enumeration

    def _dir(self, path):
        files = self.__webhdfs.list_dir(path)
        for f in files['FileStatuses']['FileStatus']:
            yield FileSystemFileDesc(
                fs=self,
                path=posixpath.join(path, f['pathSuffix']),
                name=f['pathSuffix'],
                type=f['type'].lower(),
                bytes=f['length'],
                owner=f['owner'],
                group=f['group'],
                last_modified=f['modificationTime'],
                last_accessed=f['accessTime'],
                permissions=f['permission'],
                replication=f['replication']
            )

    def _mkdir(self, path, recursive, exist_ok):
        if not recursive and not self._isdir(self.path_basename(path)):
            raise IOError("No parent directory found for {}.".format(path))
        if not exist_ok and self._exists(path):
            raise IOError("Path already exists at {}.".format(path))
        self.__webhdfs.make_dir(path)

    def _remove(self, path, recursive):
        self.__webhdfs.delete_file_dir(path, recursive)

    # File handling

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        if not self.isfile(path):
            raise FileNotFoundError("File `{}` does not exist.".format(path))

        read = self.__webhdfs.read_file(path, offset=offset, length='null' if size < 0 else size)
        if not binary:
            read = read.decode('utf-8')
        return read

    def _file_append_(self, path, s, binary):
        return self.__webhdfs.append_file(path, s)

    def _file_write_(self, path, s, binary):
        return self.__webhdfs.create_file(path, s, overwrite=True)



import six
import yaml

from omniduct.duct import Duct
from omniduct.errors import DuctNotFound, DuctProtocolUnknown
from omniduct.utils.debug import logger
from omniduct.utils.magics import MagicsProvider
from omniduct.utils.proxies import NestedDictObjectProxy


class DuctRegistry(object):
    """
    A convenient registry for `Duct` instances.

    This class provides a simple interface to a pool of configured services,
    allowing convenient lookups of available services and the creation of new
    ones. It also allows for the batch creation of services from a shared
    configuration, which is especially useful in a company deployment.
    """

    class ServicesProxy(NestedDictObjectProxy):
        """
        A wrapper around `NestedDictObjectProxy` which is used to expose the
        services attached to a `DuctRegistry` as attributes on an object,
        optionally nested by service type.
        """

        def __init__(self, registry, by_kind=True):
            self._self_registry = registry
            get_nesting = None
            if by_kind:
                def get_nesting(k, v):
                    nesting = k.split('/')
                    if v.DUCT_TYPE is not None:
                        nesting.insert(0, v.DUCT_TYPE.value)
                    return nesting

            NestedDictObjectProxy.__init__(self, registry._registry, is_flat=True, get_nesting=get_nesting)

        @property
        def registry(self):  # This will only appear at top level of proxy, since children will not be of this type
            """DuctRegistry: The registry which hosts the services."""
            return self._self_registry

        def __dir__(self):
            return NestedDictObjectProxy.__dir__(self) + ['registry']

    def __init__(self, config=None):
        """
        Args:
            config (iterable, dict, str, None): Refer to `.import_from_config`
                for more details (default: `None`).
        """
        self._registry = {}

        if config:
            self.register_from_config(config)

    def __repr__(self):
        return "<DuctRegistry with {} registered ducts>".format(len(self._registry))

    # Registration methods
    def register(self, duct, name=None, override=False, register_magics=True):
        """
        Register an existing Duct instance into the registry.

        Names of ducts can consist of any valid Python identifier, and multiple
        names can be provided as a comma separated list in which case the names
        will be aliases referring to the same `Duct` instance. Keep in mind that
        any name must uniquely identify one `Duct` instance.

        Args:
            duct (Duct): The `Duct` instance to be registered.
            name (str): An optional name to use when registering. If not
                provided this will fall back to `duct.name`. If neither is
                configured, an error will be thrown. Name can be a
                comma-separated list of names, in which case the names are
                aliases and will point to the same `Duct` instance.
            override (bool): Whether to override any existing `Duct` instance
                of the same name. If `False`, any overrides will result in an
                exception.

        Returns:
            Duct: The `Duct` instance being registered.
        """
        name = name or duct.name
        if name is None:
            raise ValueError("`Duct` instances must be named to be registered. Please either specify a name to this method call, or add a name to the Duct using `duct.name = '...'`.")
        names = [n.strip() for n in name.split(',')]
        for name in names:
            if name in self._registry and not override:
                raise ValueError("`Duct` with the same name ('{}') already present in the registry. Please pass `override=True` if you want to override the existing instance, or `name='...'` to specify a new name.".format(name))
            if register_magics and isinstance(duct, MagicsProvider):
                duct.register_magics(base_name=name)
            self._registry[name] = duct
        return duct

    def new(self, name, protocol, override=False, register_magics=True, **kwargs):
        """
        Create a new service and register it into the registry.

        Args:
            name (str): The name (or names) of the target service. If multiple
                aliases are to be used, names should be a comma separated list.
                See `.register` for more details.
            protocol (str): The protocol of the new service.
            override (bool): Whether to override any existing `Duct` instance
                of the same name. If `False`, any overrides will result in an
                exception.
            register_magics (bool): Whether to register the magics if running in
                and IPython session (default: `True`).
            **kwargs (dict): Additional arguments to pass to the constructor of
                the class associated with the nominated protocol.

        Returns:
            Duct: The `Duct` instance registered into the registry.
        """
        return self.register(
            Duct.for_protocol(protocol)(
                name=name.split(',')[0].strip(),
                registry=self,
                **kwargs
            ),
            name=name,
            override=override,
            register_magics=register_magics
        )

    # Inspection and retrieval methods
    @property
    def names(self):
        """list: The names of all ducts in the registry."""
        return sorted(self._registry.keys())

    def __getitem__(self, name):
        return self._registry[name]

    def __contains__(self, name):
        return name in self._registry

    def lookup(self, name, kind=None):
        """
        Look up an existing registered `Duct` by name and (optionally) kind.

        Args:
            name (str): The name of the `Duct` instance.
            kind (str, Duct.Type): The kind of `Duct` to which the lookup should
                be restricted.

        Returns:
            `Duct`: The looked up `Duct` instance.

        Raises:
            DuctNotFound: If no `Duct` can be found for requested name and/or
                type.
        """
        if kind and not isinstance(kind, Duct.Type):
            kind = Duct.Type(kind)
        if name not in self._registry:
            raise DuctNotFound(name)
        duct = self._registry[name]
        if kind and duct.DUCT_TYPE != kind:
            raise DuctNotFound("Duct named '{}' exists, but is not of kind '{}'.".format(name, kind.value))
        return duct

    # Exposing `Duct` instances.
    def populate_namespace(self, namespace=None, names=None, kinds=None):
        """
        Populate a nominated namespace with references to a subset of ducts.

        While a registry object is a great way to store and configure `Duct`
        instances, it is sometimes desirable to surface frequently used
        instances in other more convenient namespaces (such as the globals of
        your module).

        Args:
            namespace (dict, None): The namespace to populate. If using from a
                module you can pass `globals()`. If `None`, a new dictionary is
                created, populated and then returned.
            names (list<str>, None): The names to include in the population. If
                not specified then all names will be exported.
            kinds (list<str>, None): The kinds of ducts to include in the
                population. If not specified, all kinds will be exported.

        Returns:
            dict: The populated namespace.
        """
        if namespace is None:
            namespace = {}
        if kinds is not None:
            kinds = [Duct.Type(kind) if not isinstance(kind, Duct.Type) else kind for kind in kinds]
        for name, duct in self._registry.items():
            if (kinds is None or duct.DUCT_TYPE in kinds) and (names is None or name in names):
                namespace[name.split('/')[-1]] = duct
        return namespace

    def get_proxy(self, by_kind=True):
        """
        Return a structured proxy object for easy exploration of services.

        This method returns a proxy object to the registry upon which the `Duct`
        instances are available as attributes. This object is
        also by default structured such that one first accesses an attribute
        associated with a kind, which makes larger collections of services
        more easily navigatable.

        For example, if you have `DatabaseClient` subclass registered as
        'my_service', you could access it on the proxy using:
        >>> proxy = registry.get_proxy(by_kind=True)
        >>> proxy.databases.my_service

        Args:
            by_kind (bool): Whether to nest proxy of `Duct` instances by kind.

        Returns:
            ServicesProxy: The proxy object.
        """
        return DuctRegistry.ServicesProxy(self, by_kind=by_kind)

    # Batch registration of duct configurations
    def register_from_config(self, config):
        """
        Register a collection of Duct service configurations.

        The configuration format must be one of the following:
        - An iterable sequence of dictionaries containing a mapping between the
          keyword arguments required to instantiate the `Duct` subclass.
        - A dictionary mapping names of `Duct` instances to dictionaries of
          keyword arguments.
        - A dictionary mapping Duct types ('databases', 'filesystems', etc) to
          mappings like those immediately above.
        - A string YAML representation of one of the above (with at least one
          newline character).
        - A string filename containing such a YAML representation.

        There are three special keyword arguments that are required by the
        `DuctRegistry` instance:
        - name: Should be present only in the configuration dictionary when
          config is provided as an iterable sequence of dictionaries.
        - protocol: Which specifies which `Duct` subclass to fetch. Failure to
          correctly set this will result in a warning and an ignoring of this
          configuration.
        - register_magics (optional): A boolean flag indicating whether to
          register any magics defined by this Duct class (default: True).

        Args:
            config (iterable, dict, str, None): A configuration specified in one
                of the above described formats.
        """
        config = self._process_config(config)

        for duct_config in config:
            names = duct_config.pop('name')
            protocol = duct_config.pop('protocol')
            register_magics = duct_config.pop('register_magics', True)
            try:
                self.new(names, protocol, register_magics=register_magics, **duct_config)
            except DuctProtocolUnknown as e:
                logger.error("Failed to configure `Duct` instance(s) '{}'. {}".format("', '".join(names.split(',')), str(e)))

        return self

    def _process_config(self, config):
        """
        Extract config from file (if necessary), and coerce the configuration
        format into a generator of dictionaries of keyword arguments.
        """
        # Extract configuration from a file if necessary
        if isinstance(config, six.string_types):
            if '\n' in config:
                config = yaml.load(config)
            else:
                with open(config) as f:
                    config = yaml.load(f.read())

        if not isinstance(config, (list, dict)):
            raise ValueError("Invalid configuration detected.")

        if isinstance(config, dict):
            def max_consistent_depth(d):
                depth = 0
                if isinstance(d, dict) and d:
                    depths = []
                    for value in d.values():
                        depths.append(max_consistent_depth(value) + 1)
                    depth = max(depth, min(depths))
                return depth

            depth = max_consistent_depth(config)

            if depth == 2:
                for name, kwargs in config.items():
                    kwargs = kwargs.copy()
                    kwargs['name'] = name
                    yield kwargs
            elif depth == 3:
                for subsection in config.values():
                    for name, kwargs in subsection.items():
                        kwargs = kwargs.copy()
                        kwargs['name'] = name
                        yield kwargs
            else:
                raise ValueError("Invalid configuration detected.")

        else:
            for kwargs in config:
                if not isinstance(kwargs, dict):
                    raise ValueError("Invalid configuration detected.")
                yield kwargs



import os
import sys

__all__ = ['__author__', '__author_email__', '__version__', '__logo__', '__docs_url__']

__author__ = "Matthew Wardrop, Dan Frank"
__author_email__ = "mpwardrop@gmail.com, danfrankj@gmail.com"
__version__ = "1.0.1"
__logo__ = os.path.join(os.path.dirname(__file__), 'logo.png') if '__file__' in globals() else None
__docs_url__ = "https://omniduct.readthedocs.io/"


# These are the core dependencies, and should not include those which are used only in handling specific protocols.
# Order matters since installation happens from the end of the list
__dependencies__ = [
    "future",  # Python 2/3 support
    "six",  # Python 2/3 support
    "enum34",  # Python 3.4+ style enums in older versions of python

    "pyyaml",  # YAML configuration parsing
    "decorator",  # Decorators used by caching and documentation routines
    "progressbar2>=3.30.0",  # Support for progressbars in logging routines
    "wrapt",  # Object proxying for conveniently exposing ducts in registry

    # Database querying libraries
    "jinja2",  # Templating support in databases
    "pandas>=0.17.1",  # Various results including database queries are returned as pandas dataframes
    "sqlparse",  # Neatening of SQL based queries (mainly to avoid missing the cache)
    "sqlalchemy",  # Various integration endpoints in the database stack
]

PY2 = sys.version_info[0] == 2
if os.name == 'posix' and PY2:
    __dependencies__.append('subprocess32')  # Python 3.2+ subprocess handling for Python 2

__optional_dependencies__ = {
    # Databases
    'druid': [
        'pydruid>=0.4.0',  # Primary client
    ],

    'hiveserver2': [
        'pyhive[hive]>=0.4',  # Primary client
        'thrift>=0.10.0',  # Thrift dependency which seems not to be installed with upstream deps
    ],

    'presto': [
        'pyhive[presto]>=0.4',  # Primary client
        'werkzeug',  # Schema traversal
    ],

    'pyspark': [
        'pyspark',  # Primary client
    ],

    'snowflake': [
        'snowflake-sqlalchemy',
    ],

    # Filesystems
    'webhdfs': [
        'pywebhdfs',  # Primary client
        'requests',  # For rerouting redirect queries to our port-forwarded services
    ],

    's3': [
        'boto3',  # AWS client library
    ],

    # Remotes
    'ssh': [
        'pexpect',  # Command line handling (including smartcard activation)
    ],

    'ssh_paramiko': [
        'paramiko',  # Primary client
        'pexpect',  # Command line handling (including smartcard activation)
    ],

    # Rest clients
    'rest': [
        'requests',  # Library to handle underlying REST queries
    ],

    # Documentation requirements
    'docs': [
        'sphinx',  # The documentation engine
        'sphinx_autobuild',  # A Sphinx plugin used during development of docs
        'sphinx_rtd_theme',  # The Spinx theme used by the docs
    ],

    'test': [
        'nose',      # test runner
        'mock',      # mocking
        'pyfakefs',  # mock filesystem
        'coverage',  # test coverage monitoring
        'flake8',    # Code linting
    ]
}
__optional_dependencies__['all'] = [dep for deps in __optional_dependencies__.values() for dep in deps]
